package go_compiler

import (
	"os"
	"os/exec"
	"runtime"
)

const TempOutputDir = "tempOutput/"
const fileName = "main.go"

type GoCompiler struct{}

// NewGoCompiler creates a new GoCompiler
func NewGoCompiler() *GoCompiler {
	return &GoCompiler{}
}

// OS Operating system type
type OS = string

// Platform enums
const (
	Windows OS = "windows"
	Linux   OS = "linux"
	MacOS   OS = "darwin"
)


// MakeCommand creates a command based on the runtime platform
func MakeCommand(cmd string) *exec.Cmd {
	platform := runtime.GOOS
	switch platform {
	case Windows:
		return exec.Command("cmd", "/c", cmd)
	case Linux, MacOS:
		return exec.Command("bash", "-c", cmd)
	default:
		panic("Unsupported platform")
	}
}

// SetupTempFolders creates the temp output directory for compiled files, panics if it fails
func SetupTempFolders(tempOutputDir string) {
	// 0777 are the permissions for the directory, everyone can read, write and execute
	err := os.MkdirAll(tempOutputDir, os.ModePerm)
	if err != nil {
		panic("Error creating temp output directory:\n\n" + err.Error())
	}
}

// RemoveTempFolders removes the temp output directory for compiled files, panics if it fails
func RemoveTempFolders(tempOutputDir string) {
	err := os.RemoveAll(tempOutputDir)
	if err != nil {
		panic("Error removing temp output directory:\n\n" + err.Error())
	}
}



// CheckCompileErrors takes Go source code and checks for compile errors.
//
// The dependencies are handled automatically by go mod and go tidy.
//
// NOTE: Make sure you have an up-to-date Go installed on the system
//
// Returns the output of the compilation and an error if any
func (gb *GoCompiler) CheckCompileErrors(srcCode []byte) ([]byte, error) {
	// Make temp folders
	SetupTempFolders(TempOutputDir)
	defer RemoveTempFolders(TempOutputDir)

	// Write code to file
	err := os.WriteFile(TempOutputDir+fileName, srcCode, 0644)
	if err != nil {
		return nil, err
	}
	// Init go mod and tidy
	cmdString := "go mod init tempOutput && go mod tidy "

	// Run go build
	cmdString += " && go build -o main " + fileName

	// Run tests
	cmdString += " && go test " + fileName

	cmd := MakeCommand(cmdString)
	cmd.Dir = TempOutputDir
	return cmd.CombinedOutput()
}
